/*
 *  Measure Thread locality Penalty
 *
 *  Benchmark accessing data in various localities by threads.
 *
 *  Syntax:
 *   Unix
 *     cc -o thread_locality thread_locality.c -lpthread -lm
 *
 *   Windows - make sure WIN32 is defined.
 *
 *  Author:  Dennis Lang 2008
 */

/*
Sample runs on Windows XP Prof, HP E5440@2.83GHz 4 dual-core Intel Xeon

NOTE:  That with distance set to '0', the cache line is NOT LOCKed and the
resultant shared incrementing of the same location results in a random final
value. The total program speed varies on the counter distance with no change
after the distance is 13 fields (4*13=52 bytes).
                                               
Total time: 11.657619 seconds, distance:0, iterations:400000000, threads:4
 0:  11.577848 511924658
 1:  11.352532 490664940
 2:  11.657260 519283796
 3:  11.651834 517222244

Total time: 11.608698 seconds, distance:1, iterations:400000000, threads:4
 0:  11.449553 400000000
 1:  10.878310 400000000
 2:  11.608370 400000000
 3:  10.559528 400000000

Total time: 11.657619 seconds, distance:0, iterations:400000000, threads:4
Total time: 11.608698 seconds, distance:1, iterations:400000000, threads:4
Total time: 20.127997 seconds, distance:2, iterations:400000000, threads:4
Total time: 15.124538 seconds, distance:3, iterations:400000000, threads:4
Total time: 5.492766  seconds, distance:4, iterations:400000000, threads:4
Total time: 5.555509  seconds, distance:5, iterations:400000000, threads:4
Total time: 14.724832 seconds, distance:6, iterations:400000000, threads:4
Total time: 7.718066  seconds, distance:7, iterations:400000000, threads:4
Total time: 3.820296  seconds, distance:8, iterations:400000000, threads:4
Total time: 3.884410  seconds, distance:9, iterations:400000000, threads:4
Total time: 4.164538  seconds, distance:10, iterations:400000000, threads:4
Total time: 11.788299 seconds, distance:11, iterations:400000000, threads:4
Total time: 11.884536 seconds, distance:11, iterations:400000000, threads:4
Total time: 4.778139  seconds, distance:12, iterations:400000000, threads:4
Total time: 1.046083  seconds, distance:13, iterations:400000000, threads:4
Total time: 1.046171  seconds, distance:15, iterations:400000000, threads:4
Total time: 1.046150  seconds, distance:16, iterations:400000000, threads:4



*/
#ifdef WIN32

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
typedef HANDLE Thread_t;
typedef DWORD  Thread_return_t;
#define pthread_exit return

#else

#include <pthread.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/resource.h>
typedef pthread_t Thread_t;
typedef void      Thread_return_t;
#define WINAPI 

#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

#define USECSPERSEC 1000000
#define pthread_attr_default NULL
#define MAX_THREADS 50

unsigned long	g_distance = 0;
unsigned long	g_endDistance = 0;
unsigned long	g_stepDistance = 0;
unsigned long	g_iteration_count = 400000000;
int				g_thread_count = 4;

typedef unsigned long Counter_t;

typedef unsigned long	Counter_t;
Counter_t *				g_pCounter = 0;; 
volatile Counter_t		g_value[MAX_THREADS];
volatile double			g_elapsed[MAX_THREADS];
volatile unsigned char  g_apicId[MAX_THREADS];

char help_msg[] =
"  Thread locality tester  " __DATE__ "\n"
"\n"
" Options:\n"
"  -t <num_threads>	          ; number of threads, default %d\n"
"  -i <iterations>	          ; number of iterations, default %u\n"
"  -d <distance>,<end>,<step> ; distance between data members, default %u\n"
"  -h or -? 		          ; this help messages \n"
"\n";


#ifndef HAVE_GETOPT
/*
 * get option letter from argument vector
 */
/* See lib/libc/gen/common/optind.c for next 3 definitions. */
char *optarg = NULL; /* argument associated with option */
int opterr   = 0; /* if error message should be printed */
int optind   = 1; /* index into parent argv vector */
int optopt   = 0; /* character checked for validity */


#define BADCH (int)'?'
#define EMSG ""

/* ------------------------------------------------------------------------- */
int getopt(int nargc, char *const nargv[], const char *ostr)
{
	static char *place = EMSG; /* option letter processing */
	const char *oli; /* option letter list index */
	char *p;
	
	if (!*place) { /* update scanning pointer */
		if (optind >= nargc || *(place = nargv[optind]) != '-') {
			place = EMSG;
			return (EOF);
		}
		if (place[1] && *++place == '-') { /* found "--" */
			++optind;
			place = EMSG;
			return (EOF);
		}
	} /* option letter okay? */

	if ((optopt = (int)*place++) == (int)':' ||
		!(oli = strchr(ostr, optopt))) {
		
		/*
		* For backwards compatibility: don't treat '-' as an
		* option letter unless caller explicitly asked for it.
		*/
		if (optopt == (int)'-')
			return (EOF);
		if (!*place)
			++optind;
		if (opterr) {
			if (!(p = strrchr(*nargv, '/')))
				p = *nargv;
			else
				++p;
			fprintf(stderr, "%s: illegal option -- %c\n", p, optopt);
		}
		return (BADCH);
	}

	if (*++oli != ':') { /* don't need argument */
		optarg = NULL;
		if (!*place)
			++optind;
	} else { /* need an argument */
		if (*place) /* no white space */
			optarg = place;
		else if (nargc <= ++optind) { /* no arg */
			place = EMSG;
			if (!(p = strrchr(*nargv, '/')))
				p = *nargv;
			else
				++p;
			if (opterr)
				fprintf(stderr, "%s: option requires an argument -- %c\n",
				p, optopt);
			return (BADCH);
		} else /* white space */
			optarg = nargv[optind];
		place = EMSG;
		++optind;
	}

	return (optopt); /* dump back option letter */
}
#endif

#ifdef WIN32
typedef __int64 Tics_t;
#else
typedef struct timeval Tics_tl
#endif

/* ------------------------------------------------------------------------- */
void get_time(Tics_t * pTics)
{
#ifdef WIN32
    QueryPerformanceCounter((LARGE_INTEGER*)pTics);
#else
    gettimeofday(pTics, NULL);
#endif
}

/* ------------------------------------------------------------------------ */
double secondsElapsed(Tics_t end, Tics_t start)
{
#ifdef WIN32
    LARGE_INTEGER freq;
	QueryPerformanceFrequency(&freq);
	return (double)(end - start) / freq.QuadPart;
#else
    Tics_t delta;

	delta.tv_sec  = end.tv_sec - start.tv_sec;
    delta.tv_usec = end.tv_usec - start.tv_usec;

    if (delta.tv_usec < 0)
    {
        delta.tv_sec--;
        delta.tv_usec += USECSPERSEC;
    }

    return delta.tv_sec + (double)delta.tv_usec/1e6;
#endif
}



#define INITIAL_APIC_ID_BITS  0xFF000000  // EBX[31:24] Bits 24-31 (8 bits) return the 8-bit unique 
                                          // initial APIC ID for the processor this code is running on.
                                          

/* ------------------------------------------------------------------------- */
unsigned char GetAPIC_ID(void)
{
	unsigned int Regebx = 0;
#ifdef WIN32
	__asm
	{
		mov eax, 1
		cpuid
		mov Regebx, ebx
	}
#else
    asm
	(
		"movl $1, %%eax\n\t"	
		"cpuid"
		: "=b" (Regebx) 
		:
		: "%eax","%ecx","%edx" 
	);
#endif                                

	return (unsigned char) ((Regebx & INITIAL_APIC_ID_BITS) >> 24);

}

/* ------------------------------------------------------------------------- */
Thread_return_t WINAPI run_test(void* who_data)
{
    int who = (int)who_data;
	volatile Counter_t& counter = g_pCounter[who * (g_distance)];
	Counter_t finalCounter;

    unsigned long  i;
    unsigned long total_iterations = g_iteration_count;
    Tics_t start, end;
	double elapsed;

    /*
     * Run test
     */
    get_time(&start);

    for (i = 0; i < total_iterations; i++) 
    {
		// Increment counter which may be shared by other threads or maybe
		// 'distance' fields apart.
		counter++;
		// Exercise counter to slow down the loop some.
		if ((counter & 0x5555) == counter << 2)
			counter++;

		finalCounter = counter;
    }

    get_time(&end);

    elapsed = secondsElapsed(end, start);
 
	g_value[who] = finalCounter;
	g_elapsed[who] = elapsed;
    g_apicId[who] = GetAPIC_ID();
	pthread_exit(0);
}

/* ------------------------------------------------------------------------- */
void thread_test(bool setThreadAffinity, WORD processorCount)
{
    Tics_t start, end;
	double delta;

	int i;
	Thread_t thread[MAX_THREADS];

	// Allocate memory region where threads will update their counters.
	// User defines how far apart the counters sit in memory.
	// Distance is in units of sizeof(Counter_t).
	// Force allocation to a multiple of 4096 because it feels right.
    unsigned long request_size = (g_distance+1) * sizeof(Counter_t) * MAX_THREADS + 4;
	request_size = (request_size | 0xfff) + 1;	// force to multiples of 4096

	g_pCounter = (Counter_t*)malloc(request_size);
	memset(g_pCounter, 0, request_size);

	// Get a pointer which starts on a 4byte boundary.
	g_pCounter = (Counter_t*)((unsigned int)g_pCounter | 0x3) + 1;	// force to 4 byte boundary

    /*
     * Invoke the tests
     */
    get_time(&start);
    for (i=0; i < g_thread_count; i++)
	{
#ifdef WIN32
		DWORD thread_id;
         if ((thread[i] = CreateThread(NULL, 0, run_test, (PVOID)i, 0, &thread_id)) == NULL)
#else
        if (pthread_create(&(thread[i]), pthread_attr_default,
                    (void *) &run_test, (void*)i ))
#endif
		{
            printf("failed.\n");
		}
#ifdef WIN32
        else
        {
            if (setThreadAffinity)
            {
                DWORD threadAffinityMask = 1 << (i % processorCount);
                DWORD oldThreadAffinityMask = SetThreadAffinityMask(thread[i], threadAffinityMask);
            }
        }
#endif
	}

    /*
     * Wait for tests to finish
     */
    for (i=0; i < g_thread_count; i++)
    {
#ifdef WIN32
		WaitForSingleObject(thread[i], INFINITE);
#else
        pthread_join(thread[i], NULL);
#endif
    }

	get_time(&end);
	delta = secondsElapsed(end, start);

    printf("Total time: %.6f seconds, distance:%u, iterations:%u, threads:%u\n",
        delta, g_distance, g_iteration_count, g_thread_count);

	for (i=0; i < g_thread_count; i++)
	{
		printf("%2d: %10.6f %8u %04x\n",
			i, 
			g_elapsed[i], 
			g_value[i],
            g_apicId[i]);
	}

    return;
}


/* ------------------------------------------------------------------------- */
int main(int argc, char *argv[])
{
    bool setThreadAffinity = false;

    int  c;
    char* nextPtr;

    /*
     * Parse our arguments
     */
    while ((c = getopt(argc, argv, "at:d:i:n:h?")) != -1)
    {
        switch (c)
        {
          case 'a':   // set thread affinity
              setThreadAffinity = true;
              break;

          case 'd':
            /* distance between counters */
            g_distance = strtoul(optarg, &nextPtr, 10);
            if (nextPtr >= optarg)
            {
                optarg = nextPtr+1;
                g_endDistance = strtoul(optarg, &nextPtr, 10);
                if (nextPtr >= optarg)
                {
                    optarg = nextPtr+1;
                    g_stepDistance = strtoul(optarg, &nextPtr, 10);
                    if (optarg == nextPtr)
                        g_stepDistance = 0;
                }
                else
                    g_endDistance = 0;
            }
            else
                g_distance = 0;

            break;

          case 'i': // set # iterations
          case 'n':
            /* size and iteration count were specified; others default */
            g_iteration_count = strtoul(optarg, NULL, 10);
            break;

          case 't': // set # threads
            /* size, iteration count, and thread count were specified */
            g_thread_count = atoi(optarg);
            if (g_thread_count > MAX_THREADS) g_thread_count = MAX_THREADS;
            break;

          case 'h': // show help message
          case '?':
            printf(help_msg, g_thread_count, g_iteration_count, g_distance);
			exit(0);

          default:
            printf("Unrecognized arguments: %c\n", c);
            exit(1);
        }
    }

    WORD processorCount = 0;
#ifdef WIN32
    DWORD procAffinityMask = 0;
    DWORD sysAffinityMask = 0;
    GetProcessAffinityMask(GetCurrentProcess(), &procAffinityMask, &sysAffinityMask);
    for (WORD i = 0; i < 16; i++)
    {
        if ((sysAffinityMask & (1 << i)) != 0)
            processorCount++;
    }
#endif

	printf("Measure thread locality Penalty\n"
        "  Distance   = %u, End:%u, Step:%u\n"
        "  Iterations = %u\n"
        "  Threads    = %u\n"
        "  Processors = %u\n"
        "  %s\n",
        g_distance, g_endDistance, g_stepDistance, 
        g_iteration_count, g_thread_count,
        processorCount,
        setThreadAffinity ? "Set Thread Affinity": "");
	
    // Run test - optionally run mutliple times if endDistance and step set.
    if (g_endDistance > g_distance && g_stepDistance > 0)
    {
        do
        {
	        thread_test(setThreadAffinity, processorCount);
            g_distance += g_stepDistance;
        } while (g_distance <= g_endDistance);
    }
    else
	    thread_test(setThreadAffinity, processorCount);

    return 0;
}

